// Copyright 1995 SF Bellenot
// All Rights Reserved
// --yes it is overkill: bellenot@math.fsu.edu
// Populate log surface net nodes at points X, Y, Z

#include <math.h>
#include "bench.h"

const double PI = M_PI;

Real SinArray[MAX_COLS];
Real CosArray[MAX_COLS];
int ArraysFull;

void FillTrigArrays()
{
	double fTmp = ( 2.0 * PI ) / MAX_COLS;

	SinArray[MAX_COLS/2] = SinArray[0] = 0.0;
	SinArray[3*MAX_COLS/4] = -1.0;
	SinArray[MAX_COLS/4] = 1.0;

	CosArray[3*MAX_COLS/4] = CosArray[MAX_COLS/4] = 0.0;
	CosArray[MAX_COLS/2] = -1.0;
	CosArray[0] = 1.0;

	for ( int i = 1; i < MAX_COLS/4; i++ )
	{
		SinArray[MAX_COLS/2 - i] = SinArray[i] = sin ( i * fTmp );
		SinArray[MAX_COLS/2 + i] = SinArray[MAX_COLS - i] = - sin ( i * fTmp );
		CosArray[MAX_COLS - i] = CosArray[i] = cos ( i * fTmp );
		CosArray[MAX_COLS/2 - i] = CosArray[MAX_COLS/2 + i] = -cos ( i * fTmp );
	}
}

int BuildLogSurface( x32_LogSurf * Data, Real SmallDiam, Real LargeDiam,
	int MaxRow )
{
	int i, j;
	double Radius;
	double DiamTaper = ( LargeDiam - SmallDiam ) / MaxRow;

	if ( ! ArraysFull )
	{
		ArraysFull = 1;
		FillTrigArrays();
	}

	for ( i = 0; i <= MaxRow; i++ )    /* for each slice */
	{
		Radius = ( LargeDiam -  DiamTaper * i )/2;
		for ( j = 0; j < MAX_COLS; j++ )
		{
			// now centered at zero
			(Data[i])[j].x = Radius * CosArray[j];
			// resting on x-axis
			(Data[i])[j].z = Radius + Radius * SinArray[j];
		}
	}

	return 1;
}
