/*
 * Copyright 1995 SF Bellenot HF Qualls
 * All Rights Reserved
 * --yes it is overkill: bellenot@math.fsu.edu
 * fourier.c
 * bench mark for scanner code based on build32.c++
 * Authors: Steve Bellenot, PhD., H. Qualls
 */

#include <math.h>
typedef struct
{
	short iX,iY;
} POINT_I;

typedef struct
{
	float fX, fY;
} POINT_F;

#if 0
const int ciMAXPOINTS = 120;
const int ciMAXFUNCS = 7;
#else
#define ciMAXPOINTS 120
#define ciMAXFUNCS 7
#endif

struct DataEveryThing
{
	int Valid;
	double X;
	double Y;
	double R;
	double Theta;
	double Error;
	double Basis[ciMAXFUNCS]; 
}
Data[ciMAXPOINTS];

double MaxY;
double MaxX;
double MinX;
double OffsetY;
double OffsetX;
double MinTheta;
double MaxTheta;

double Basis[ciMAXFUNCS][ciMAXPOINTS];
double Ortho[ciMAXFUNCS][ciMAXFUNCS];
double Coeff[ciMAXFUNCS];

int DataNum;
int FunNum;

int NumStartValidData;
int NumEndValidData;
float ValidPerCent;

char * prefix = "b32";
const int buf_len = 100;

double basis0 ( void    ) { return 1.0; }
double basis1 ( double x) { return sin(x); }
double basis2 ( double x) { return cos(x); }
double basis3 ( double x) { return sin(2*x); }
double basis4 ( double x) { return cos(2*x); }
double basis5 ( double x) { return sin(3*x); }
double basis6 ( double x) { return cos(3*x); }
double basis7 ( double x) { return sin(4*x); }
double basis8 ( double x) { return cos(4*x); }

double basis0p ( void    ) { return 0.0; }
double basis1p ( double x) { return cos(x); }
double basis2p ( double x) { return -sin(x); }
double basis3p ( double x) { return 2.0*cos(2*x); }
double basis4p ( double x) { return -2.0*sin(2*x); }
double basis5p ( double x) { return 3.0*cos(3*x); }
double basis6p ( double x) { return -3.0*sin(3*x); }
double basis7p ( double x) { return 4.0*cos(4*x); }
double basis8p ( double x) { return -4.0*sin(4*x); }

double basis0pp ( void    ) { return 0.0; }
double basis1pp ( double x) { return -sin(x); }
double basis2pp ( double x) { return -cos(x); }
double basis3pp ( double x) { return -4.0*sin(2*x); }
double basis4pp ( double x) { return -4.0*cos(2*x); }
double basis5pp ( double x) { return -9.0*sin(3*x); }
double basis6pp ( double x) { return -9.0*cos(3*x); }
double basis7pp ( double x) { return -16.0*sin(4*x); }
double basis8pp ( double x) { return -16.0*cos(4*x); }

char * basis0str = "1.0";
char * basis1str = "sin(x)";
char * basis2str = "cos(x)";
char * basis3str = "sin(2*x)";
char * basis4str = "cos(2*x)";
char * basis5str = "sin(3*x)";
char * basis6str = "cos(3*x)";
char * basis7str = "sin(4*x)";
char * basis8str = "cos(4*x)";

typedef double (* DFunD)(double);
DFunD Executable[ciMAXFUNCS];
DFunD Prime[ciMAXFUNCS];
DFunD TwoPrime[ciMAXFUNCS];
char* Writeable[ciMAXFUNCS];

/*
 * This function populates the structure asGoodCamPtF given points arriving
 * via apsPoint[].  asGoodCamPtF is modified and the number of kept points
 * is returned.
 */

int LoadPoints( POINT_F asGoodCamPtF[], POINT_F apsPoint[] )
{
	int iI;
  int iPointCount = 0;
  for (iI = 0; iI < 120; iI++ ) {
        if ( apsPoint[iI].fY > 0.0 && fabs(apsPoint[iI].fX) < 12.0 )
		{
          asGoodCamPtF[iPointCount++] = apsPoint[iI];
        }
  }
  NumStartValidData = NumEndValidData = iPointCount;
  return iPointCount;
}

void DoAble()
{
	int j = 0;

	Executable[j] = (DFunD)(basis0);
	Prime[j] = (DFunD)(basis0p);
	TwoPrime[j] = (DFunD)(basis0pp);
	Writeable[j] = basis0str;
	j++;

	Executable[j] = basis1;
	Prime[j] = basis1p;
	TwoPrime[j] = basis1pp;
	Writeable[j] = basis1str;
	j++;

	Executable[j] = basis2;
	Prime[j] = basis2p;
	TwoPrime[j] = basis2pp;
	Writeable[j] = basis2str;
	j++;

	Executable[j] = basis3;
	Prime[j] = basis3p;
	TwoPrime[j] = basis3pp;
	Writeable[j] = basis3str;
	j++;

	Executable[j] = basis4;
	Prime[j] = basis4p;
	TwoPrime[j] = basis4pp;
	Writeable[j] = basis4str;
	j++;


  if ( j == ciMAXFUNCS )
    return;

	Executable[j] = basis5;
	Prime[j] = basis5p;
	TwoPrime[j] = basis5pp;
	Writeable[j] = basis5str;
	j++;

  if ( j == ciMAXFUNCS )
    return;

	Executable[j] = basis6;
	Prime[j] = basis6p;
	TwoPrime[j] = basis6pp;
	Writeable[j] = basis6str;
	j++;

  if ( j == ciMAXFUNCS )
    return;

	Executable[j] = basis7;
	Prime[j] = basis7p;
	TwoPrime[j] = basis7pp;
	Writeable[j] = basis7str;
	j++;

  if ( j == ciMAXFUNCS )
    return;

	Executable[j] = basis8;
	Prime[j] = basis8p;
	TwoPrime[j] = basis8pp;
	Writeable[j] = basis8str;
	j++;

}

double guess ( int k )
{
	int i;
	double answer = 0.0;

	for ( i = 0; i < FunNum; i++ )
		answer += Coeff[i] * Data[k].Basis[i];
	return answer;
}

double compute ( double theta )
{
	int i;
	double answer = 0.0;

	for ( i = 0; i < FunNum; i++ )
		answer += Coeff[i] * Executable[i] ( theta );
	return answer;
}

double compute_prime ( double theta )
{
	int i;
	double answer = 0.0;

	for ( i = 0; i < FunNum; i++ )
		answer += Coeff[i] * Prime[i] ( theta );
	return answer;
}

double compute_twoprime ( double theta )
{
	int i;
	double answer = 0.0;

	for ( i = 0; i < FunNum; i++ )
		answer += Coeff[i] * TwoPrime[i] ( theta );
	return answer;
}

double solve_for_y ( double x, double theta )
{
/*
	//find the y-value on curve near x-value
	// uses newton's method for solutio
	// x_n+1 = x_n - f(x_n)/f'(x_n)
	// where x_0 is theta.

	// no error checking cause the function is well-behaved
	// well it can go to the wrong root!
*/
	int i;
	double error = 0.00005;

	double old = theta;
	double next;

	for ( i=0; i < 10; i++ )
	{
		double temp1 = compute ( old );
		double temp2 = compute_prime ( old );
		double c = cos ( old );
		next = old - (c * temp1 - x) / ( c * temp2 - sin ( old ) * temp1 );

		if ( fabs (next - old ) < error )
			break;
		old = next;
	}

	return sin(next)*compute(next);
}

/* works for top or bottom -- uses secant method
// which avoids the second derivative.
 */
extern int major;

double top_secant ( double theta1, double theta2 )
{
	int i;
	double error = 0.00005;
	double old1, old1F;
	double old2, old2F;
	double new_ratio, old_ratio;
	double next;
	old1 = theta1;
	old1F = sin ( old1 ) * compute_prime ( old1 )
		+ cos ( old1 ) * compute ( old1 );

	next = theta2;
	new_ratio = fabs ( old1 - next );
	for ( i = 0; i < 10; i++ )
	{
		old2 = old1;
		old2F = old1F;
		old1 = next;
		old_ratio = new_ratio;
		old1F = sin ( old1 ) * compute_prime ( old1 )
			+ cos ( old1 ) * compute ( old1 );

		next = old1 - old1F * ( old1 - old2 ) / ( old1F - old2F );
		new_ratio = fabs (next - old1 );
		#if 0
		if ( ( i > 0 ) && ( new_ratio > 2.0 * old_ratio ) )
			if ( verbose )
				cerr << "growing " << i <<" " << next
					<< " theta1 was " << theta1 << " major " << major << endl;;
		#endif
		if ( new_ratio < error )
			break;
		#if 0		
		if ( i > 5 )
			if ( verbose )
				cerr << " i is " << i << " next is " << next
					<< " theta1 was " << theta1 << " major " << major << endl;;
		#endif
	}
	return next;
}

/*
// works for top or bottom -- uses secant method
// which avoids the second derivative.
 */

double left_secant ( double theta1, double theta2 )
{
	int i;
	double error = 0.00005;
	double old1, old1F;
	double old2, old2F;
	double new_ratio, old_ratio;
	double next;
	old1 = theta1;
	old1F = cos ( old1 ) * compute_prime ( old1 )
		- sin ( old1 ) * compute ( old1 );

	next = theta2;
	new_ratio = fabs ( old1 - next );
	for ( i = 0; i < 10; i++ )
	{
		old2 = old1;
		old2F = old1F;
		old1 = next;
		old_ratio = new_ratio;
		old1F = cos ( old1 ) * compute_prime ( old1 )
			- sin ( old1 ) * compute ( old1 );

		next = old1 - old1F * ( old1 - old2 ) / ( old1F - old2F );
		new_ratio = fabs (next - old1 );
		#if 0
		if ( ( i > 0 ) && ( new_ratio > 2.0 * old_ratio ) )
			if ( verbose )
				cerr << "left growing " << i <<" " << next
					<< " theta1 was " << theta1 << " major " << major << endl;;
		#endif
		if ( new_ratio < error )
			break;
		#if 0		
		if ( i > 5 )
			if ( verbose )
				cerr << "left i is " << i << " next is " << next
					<< " theta1 was " << theta1 << " major " << major << endl;;
		#endif
	}
	return next;
}

void output ( int major, int minor );

int weed ( int major, int minor )
{
	int i, j;
	int index;
	double Limit;
	double SumErrorR;
	double SumErrorRR;
	double ErrorStdDev;
	double Threshold = 2.0;

#ifdef HISTOGRAM
	const int hist_len = 10;
	const double hist_fac = 20.0;
	int pos_hist[hist_len];
	int neg_hist[hist_len];
#endif

#if GNUPLOT
	char buf[buf_len], buf2[buf_len];
	ostrstream dataname ( buf, buf_len );
	ostrstream farname ( buf2, buf_len );
	dataname << prefix << "." << major << "." << minor << ".trash" << ends;
	farname << prefix << "." << major << "." << minor << ".far" << ends;
#endif

#ifdef HISTOGRAM
	for ( i = 0; i < hist_len; i++ )
	{
		pos_hist[i] = neg_hist[i] = 0;
	}
#endif

	SumErrorR = SumErrorRR = 0.0;
	for ( i = 0; i < DataNum; i++ )
	{
		if ( ! Data[i].Valid )
			continue;
		Data[i].Error = Data[i].R - guess ( i );
		SumErrorR += Data[i].Error;
		SumErrorRR += Data[i].Error * Data[i].Error;

#ifdef HISTOGRAM
		if ( Data[i].Error < 0.0 )
		{
			index = - hist_fac * Data[i].Error;
			index = (index < hist_len)? index : hist_len - 1;
			neg_hist[index]++;
		}
		else
		{
			index = hist_fac * Data[i].Error;
			index = (index < hist_len)? index : hist_len - 1;
			pos_hist[index]++;
		}
#endif
	}
	ErrorStdDev = sqrt ( (SumErrorRR - SumErrorR * SumErrorR / DataNum)
		/(DataNum - 1) );


#ifdef HISTOGRAM
	for ( i = hist_len - 1; i >= 0; i-- )
	{
		cout << neg_hist[i] << " ";
	}

	cout << "***" ;
	for ( i = 0; i < hist_len; i++ )
	{
		cout << pos_hist[i] << " ";
	}
#endif

	Limit = Threshold * ErrorStdDev;
	Limit = (Limit > 0.45) ? 0.45: Limit;

#if GNUPLOT
	ofstream of_data;
	ofstream of_far;
	of_data.open ( dataname.str() );
	of_far.open ( farname.str() );
#endif

	j = 0;
	for ( i = 0; i < DataNum; i++ )
	{
		if ( ! Data[i].Valid )
			continue;
#if GNUPLOT
		if ( fabs ( Data[i].Error ) > 0.1 )
		{
			of_far << Data[i].Theta << " " << Data[i].R << endl;
		}
#endif
		if ( Data[i].Error > Limit ) /* Throw it out */
		{
			j++;
			Data[i].Valid = 0;
			#ifdef DISPLAY
			POINT_F(Data[i].X, Data[i].Y).Draw(FG_RED);
			#endif
			NumEndValidData--;
#if GNUPLOT
			of_data << Data[i].Theta << " " << Data[i].R << endl;
#endif
		}
	}

#if GNUPLOT
	of_data.close();
	of_far.close();
#endif

	return ( j > 0 );
}

#if 0		
int dat_open ( char * fname )
{
	in_file.open ( fname );

	if ( ! in_file )
		return 0;
	return 1;
}
#endif

#if 0		
int input ( )
{
	int i = 0;
	int j = 0;

	if ( ! in_file )
		return DataNum = 0;

	MaxY = 0.0;
	MaxX = 0.0;
	MinX = 0.0;
	while ( i < ciMAXPOINTS )
	{
		in_file >> Data[i].X >> Data[i].Y;
		if ( ! in_file )
			break;
		if ( Data[i].Y >= 2.0 )
		{
			MaxY = ( Data[i].Y > MaxY )? Data[i].Y : MaxY;
			MaxX = ( Data[i].X > MaxX )? Data[i].X : MaxX;
			MinX = ( Data[i].X < MinX )? Data[i].X : MinX;
			Data[i].Valid = 1;
			i++;
		}
		j++;
		if ( j == 120 )
			break;
	}
	OffsetY = MaxY/1.9;
	OffsetX = (MaxX + MinX)/2.0;
	if ( OffsetX > 1.0 )
	{
		cerr << "OffsetX too far " << OffsetX << endl;
		OffsetX = 1.0;
	}
	else if ( OffsetX < -1.0 )
	{
		cerr << "OffsetX too far " << OffsetX << endl;
		OffsetX = -1.0;
	}
	if ( j != 120 )
		cerr << "error in input j is " << j << endl;
	return DataNum = i;
}
#endif
#if 0
void bottom ( int major, int minor )
{
	double MinTheta = 4.0;
	double MaxTheta = -4.0;
	int i, MinI, MaxI;
	double Radius, LLx, LLy, LRx, LRy, URx, URy, ULx, ULy;
	char buf2[buf_len];
	ostrstream filename ( buf2, buf_len );
	
	filename << prefix << "." << major << "." << minor << ".bottom" << ends;
	ofstream of_data;

	of_data.open ( filename.str() );
	for ( i = 0; i < DataNum; i++ )
	{
		if ( Data[i].Valid )
		{
			if ( ( Data[i].Theta < MinTheta ) && ( Data[i].Theta > - M_PI_2 ) )
			{
				MinTheta = Data[i].Theta;
				MinI = i;
			}
			if ( ( Data[i].Theta > MaxTheta ) && ( Data[i].Theta < - M_PI_2 ) )
			{
				MaxTheta = Data[i].Theta;
				MaxI = i;
			}
		}
	}

	Radius = guess ( MinI );
	LRx = cos ( MinTheta ) * Radius;
	LRy = sin ( MinTheta ) * Radius;
	of_data << MinTheta << " " << Radius << endl;

	Radius = guess ( MaxI );
	LLx = cos ( MaxTheta ) * Radius;
	LLy = sin ( MaxTheta ) * Radius;
	of_data << MaxTheta << " " << Radius << endl;

	double Theta = - MinTheta;
	URy = solve_for_y ( LRx, Theta );
	of_data << atan2( URy, LRx) << " " << hypot(LRx, URy) << endl;

	Theta = - MaxTheta;
	/*
	// this produced the wrong point once
	// that is ULy was LLy because -MaxTheta was nearer
	// LLy than what should have been ULy so newton found LLy again.
	*/

	ULy = solve_for_y ( LLx, Theta );
	of_data << atan2( ULy, LLx) << " " << hypot(LLx, ULy) << endl;

	double CenterX = (LRx + LLx)/2.0 + OffsetX;
	double CenterY = (LRy + LLy + URy + ULy)/4.0 + OffsetY;
	of_data << atan2( CenterY - OffsetY, CenterX - OffsetX) << " "
		<< hypot(CenterX - OffsetX, CenterY - OffsetY) << endl;

	double RightR = hypot (
		cos ( Data[MinI].Theta ) * guess ( MinI ) + OffsetX - CenterX,
		sin ( Data[MinI].Theta ) * guess ( MinI ) + OffsetY - CenterY );
	double RightTheta = atan2 (
		sin ( Data[MinI].Theta ) * guess ( MinI ) + OffsetY - CenterY,
		cos ( Data[MinI].Theta ) * guess ( MinI ) + OffsetX - CenterX );
	double LeftR = hypot (
		cos ( Data[MaxI].Theta ) * guess ( MaxI ) + OffsetX - CenterX,
		sin ( Data[MaxI].Theta ) * guess ( MaxI ) + OffsetY - CenterY );
	double LeftTheta = atan2 (
		sin ( Data[MaxI].Theta ) * guess ( MaxI ) + OffsetY - CenterY,
		cos ( Data[MaxI].Theta ) * guess ( MaxI ) + OffsetX - CenterX );
	double DTheta = (RightTheta - LeftTheta)/32.0;

	for ( i = 0; i <= 32; i++ )
	{
		/* new co-ordinates */
		double R = (i * RightR + (32 - i) * LeftR)/32.0;
		double Theta = LeftTheta + i * DTheta;
		double X = CenterX + R * cos(Theta);
		double Y = CenterY + R * sin(Theta);
		/* old co-ordinates */
		Theta = atan2 (Y - OffsetY, X - OffsetX );
		R = hypot (X - OffsetX, Y - OffsetY );
		of_data << Theta << " " << R << endl;
	}
	
	of_data.close();
}
#endif
#if 0
void corner ( int major, int minor )
{
	char buf2[buf_len];
	ostrstream filename ( buf2, buf_len );

	
	filename << prefix << "." << major << "." << minor << ".corn" << ends;
	ofstream of_data;
	of_data.open ( filename.str() );

if ( minor == 3 )
{
	double TopTheta = top_secant ( M_PI_2 - 0.1, M_PI_2 + 0.1 );
	double BotTheta = top_secant ( -M_PI_2 - 0.1, -M_PI_2 + 0.1 );
	double LeftTheta = left_secant ( M_PI - 0.1, M_PI + 0.1 );
	double RightTheta = left_secant ( - 0.1, 0.1 );
	of_data << RightTheta << " " << compute ( RightTheta ) << endl;
	of_data << TopTheta << " " << compute ( TopTheta ) << endl;
	of_data << LeftTheta << " " << compute ( LeftTheta ) << endl;
	of_data << BotTheta << " " << compute ( BotTheta ) << endl;
	
}

#if 0
	of_data << RightTheta << " " << compute ( RightTheta ) << endl;
	double DTheta = (TopTheta - RightTheta)/ 8.0;
	double Theta = RightTheta;
	for ( i = 1; i < 8; i++ )
	{
		Theta += DTheta;
		of_data << Theta << " " << compute ( Theta ) << endl;
	}

	of_data << TopTheta << " " << compute ( TopTheta ) << endl;
	DTheta = (LeftTheta - TopTheta)/ 8.0;
	Theta = TopTheta;
	for ( i = 1; i < 8; i++ )
	{
		Theta += DTheta;
		double Theta = TopTheta + i * DTheta;
		of_data << Theta << " " << compute ( Theta ) << endl;
	}

	of_data << LeftTheta << " " << compute ( LeftTheta ) << endl;
	DTheta = (BotTheta - LeftTheta + 2.0 * M_PI)/ 8.0;
	Theta = LeftTheta - 2.0 * M_PI;
	for ( i = 1; i < 8; i++ )
	{
		Theta += DTheta;
		of_data << Theta << " " << compute ( Theta ) << endl;
	}

	of_data << BotTheta << " " << compute ( BotTheta ) << endl;
	DTheta = (RightTheta - BotTheta)/ 8.0;
	Theta = BotTheta;
	for ( i = 1; i < 8; i++ )
	{
		Theta += DTheta;
		of_data << Theta << " " << compute ( Theta ) << endl;
	}
#endif
	of_data.close();
}
#endif
#if 0
void output ( int major, int minor )
{
	int i;
	char buf[buf_len], buf2[buf_len];
	ostrstream dataname ( buf, buf_len );
	ostrstream filename ( buf2, buf_len );

	//bottom ( major, minor );
	corner ( major, minor );
	dataname << prefix << "." << major << "." << minor << ".data" << ends;
	filename << prefix << "." << major << "." << minor << ends;

	ofstream of_data;
	of_data.open ( dataname.str() );
	for ( i = 0; i < DataNum; i++ )
	{
		if ( Data[i].Valid )
			of_data << Data[i].Theta << " " << Data[i].R << endl;
	}
	of_data.close();
	
	ofstream of_cmd ( filename.str() );
	of_cmd << "set polar\nset title \"" << filename.str() << "\"\n"
		<< "plot '" << dataname.str() << "', '"
		<< filename.str() << ".far', '"
		<< filename.str() << ".trash', '"
		//<< filename.str() << ".bottom', \\\n";
		<< filename.str() << ".corn', \\\n";
		//<< filename.str() << ".trash', \\\n";

	for ( i = 0; i < FunNum; i++ )
	{
		if ( i && ( Coeff[i] >= 0.0 ) )
			of_cmd << "+";
		of_cmd << Coeff[i] << "*" << Writeable[i];
	}
	of_cmd << endl;
	of_cmd.close();
}
#endif

void polarize ()
{
	int i;
	int j;

	#if 0
	//OffsetY = MaxY/2.0;
	// We need to have OffsetY a bit too large
	// This removes alot of the badness at the bottom.
	OffsetY = MaxY/1.9;
	OffsetX = (MaxX + MinX)/2.0;
	if ( OffsetX > 1.0 )
	{
		cerr << "OffsetX too far " << OffsetX << endl;
		OffsetX = 1.0;
	}
	else if ( OffsetX < -1.0 )
	{
		cerr << "OffsetX too far " << OffsetX << endl;
		OffsetX = -1.0;
	}
	#endif
	for ( i = 0; i < DataNum; i++ )
	{
		Data[i].R = hypot (Data[i].X - OffsetX, Data[i].Y - OffsetY );
		Data[i].Theta = atan2 (Data[i].Y - OffsetY, Data[i].X - OffsetX );
		for ( j = 0; j < FunNum; j++ )
			Data[i].Basis[j] = Executable[j] ( Data[i].Theta );

	}
}

void LoadFunctions ()
{
	int i, j;

	for ( i = 0; i < DataNum; i++ )
		if ( Data[i].Valid )
			for ( j = 0; j < FunNum; j++ )
				Basis[j][i] = Data[i].Basis[j];
		else
			for ( j = 0; j < FunNum; j++ )
				Basis[j][i] = 0.0;
}

void GramSchmidt ()
{
	int i, j, k;
	double Scaler;

	for ( j = 0; j < FunNum; j++ )
	{
		for ( i = 0; i < FunNum; i++ )
		{
			Ortho[j][i] = 0.0;
		}
		Ortho[j][j] = 1.0;
	}

	for ( j = 0; j < FunNum; j++ )
	{
		
		for ( k = 0; k < j; k++ )
		{
			Scaler = 0.0;
			for ( i = 0; i < DataNum; i++ )
				Scaler += Basis[k][i] * Basis[j][i];
			for ( i = 0; i < DataNum; i++ )
				Basis[j][i] -= Scaler * Basis[k][i];
			for ( i = 0; i < FunNum; i++ )
			{
				Ortho[j][i] -= Scaler * Ortho[k][i];
			}
		}

		Scaler = 0.0;
		for ( i = 0; i < DataNum; i++ )
			Scaler += Basis[j][i] * Basis[j][i];
		Scaler = sqrt(Scaler);
		for ( i = 0; i < DataNum; i++ )
			Basis[j][i] = Basis[j][i]/Scaler;
		for ( i = 0; i < FunNum; i++ )
		{
			Ortho[j][i] /= Scaler;
		}
		
	}
}

double Check[ciMAXFUNCS][ciMAXFUNCS];

#if 0		
void check ()
{
	int i, j, k;
	double Scaler;

	for ( j = 0; j < FunNum; j++ )
	{
		for ( k = 0; k <= j; k++ )
		{
			Scaler = 0.0;
			for ( i = 0; i < DataNum; i++ )
				Scaler += Basis[j][i] * Basis[k][i];
			Check[j][k] = Check[k][j] = Scaler;
		}

	}

	for ( j = 0; j < FunNum; j++ )
	{
		for ( k = 0; k < FunNum; k++ )
			cout << Check[j][k] << "	";
		cout << endl;
	}
	cout << endl;

	for ( j = 0; j < FunNum; j++ )
	{
		for ( k = 0; k < FunNum; k++ )
			cout << Ortho[j][k] << "	";
		cout << endl;
	}
	cout << endl;
}
#endif

void solve ()
{
	int i, j;
	double Scaler;

	for ( i = 0; i < FunNum; i++ )
	{
		Coeff[i] = 0.0;
	}

	for ( j = 0; j < FunNum; j++ )
	{
		Scaler = 0.0;
		for ( i = 0; i < DataNum; i++ )
		{
			Scaler += Data[i].R * Basis[j][i];
		}
		for ( i = 0; i < FunNum; i++ )
		{
			Coeff[i] += Scaler * Ortho[j][i];
		}
	}
}

#if ARC_MAIN
int main ( int argc, char ** argv )
{
	int i, j;
	int count = 1;
	int major = 1;
	char cbuff[buf_len];
	ostrstream buff ( cbuff, buf_len );

	if ( argc > 1 )
	{
		prefix = argv[1];
	}
	else 
	{
		prefix = "pipe";
	}

	buff << prefix << ".dat" << ends;

	if ( ! dat_open ( buff.str() ) )
		return 1;

	DoAble();
	FunNum = 5; // default
	while ( in_file )
	{
		input();

		if ( !DataNum )
			return 1;
//for ( int p = 0; p < 1000; p++ )
//{
		polarize();
		while ( count  < 4 ) // count differs in Build32
		{
			LoadFunctions();
			GramSchmidt();
			solve();
#if 0
			for ( i = 0; i <= j; i++ )
				cout << Coeff[i] << " ";
			cout << endl;
#endif
			output(major, count);
			if ( weed(major, count) )
				count++;
			else
				break;
		}
		major++;
		count = 1;
//}
//return 0;
	}

	return 0;
}
#endif

POINT_I s32[32];      
#if 0		
int major = 1;
void Results ( )
{
	int i;
	char buf[buf_len], buf2[buf_len];
	ostrstream dataname ( buf, buf_len );
	dataname << prefix << "." << major << ends;

	ofstream of_data;
	of_data.open ( dataname.str() );
	for ( i = 0; i < 32; i++ )
	{
		of_data << 0.001 * s32[i].iX << " "
			<< 0.001 * s32[i].iY << endl;
	}
	of_data.close();
	major++;
}
#endif

int F2I ( double x ) { return (int) ( 1000.0 * (x + 0.0005) ); }

int Fill32( void )
{
	int i, j;
	double R;
	double DTheta;
	double Theta;

#if BIAS
	// Bias code. It is easy to add or substract a bias cause it
	// is just the leading coefficient that needs to be changed
	const double cfHARDWARE_BIAS = 0.0;
	double fUserBias = 0.0;
	Coeff[0] += cfHARDWARE_BIAS + fUserBias;
#endif

	double TopTheta = top_secant ( M_PI_2 - 0.1, M_PI_2 + 0.1 );
	double BotTheta = top_secant ( -M_PI_2 - 0.1, -M_PI_2 + 0.1 );
	double LeftTheta = left_secant ( M_PI - 0.1, M_PI + 0.1 );
	double RightTheta = left_secant ( - 0.1, 0.1 );

	/* 0 is a failure */
	if ( (TopTheta > M_PI - 0.1 ) || (TopTheta < 0.1) )
		return 0;
	if ( (BotTheta > - 0.1 ) || (BotTheta < -M_PI + 0.1) )
		return 0;
	if ( (LeftTheta > M_PI + M_PI_2 - 0.1 ) || (LeftTheta < M_PI_2 + 0.1) )
		return 0;
	if ( (RightTheta > M_PI_2 - 0.1 ) || (RightTheta < -M_PI_2 + 0.1) )
		return 0;

	DTheta = (TopTheta - RightTheta)/ 8.0;
	Theta = RightTheta;
	j = 0;
	for ( i = 0; i < 8; i++ )
	{
		R = compute ( Theta );
		s32[j].iX = F2I(R * cos(Theta) + OffsetX);
		s32[j++].iY = F2I(R * sin (Theta) + OffsetY);
		Theta += DTheta;
	}

	DTheta = (LeftTheta - TopTheta)/ 8.0;
	Theta = TopTheta;
	for ( i = 0; i < 8; i++ )
	{
		R = compute ( Theta );
		s32[j].iX = F2I(R * cos(Theta) + OffsetX);
		s32[j++].iY = F2I(R * sin (Theta) + OffsetY);
		Theta += DTheta;
	}

	DTheta = (BotTheta - LeftTheta + 2.0 * M_PI)/ 8.0;
	Theta = LeftTheta - 2.0 * M_PI;
	for ( i = 0; i < 8; i++ )
	{
		R = compute ( Theta );
		s32[j].iX = F2I(R * cos(Theta) + OffsetX);
		s32[j++].iY = F2I(R * sin (Theta) + OffsetY);
		Theta += DTheta;
	}

	DTheta = (RightTheta - BotTheta)/ 8.0;
	Theta = BotTheta;
	for ( i = 0; i < 8; i++ )
	{
		R = compute ( Theta );
		s32[j].iX = F2I(R * cos(Theta) + OffsetX);
		s32[j++].iY = F2I(R * sin (Theta) + OffsetY);
		Theta += DTheta;
	}
	return -1; /*OK*/
}

/*/ needs to be called before Build32*/
void InitBuild32( void )
{
	DoAble();
}

/*
// Near the original, InArray contains ArrayCount data elements to process
// The output is put into the POINT_I s32[32] array as a side effect
// Need to include a basis, we can do this by just changing the the
// Coefficient of the constant basis vector
*/

int Build32( POINT_F InArray[], int ArrayCount,
	int harmonic, int retrys )
	/*
//defaults int harmonic = 2, int retrys = 2 )
	*/
{
	int i, j;
	float cfCNTR_LIM = 2.0;
	int major = 0;
	int count = 0;
/*
	// harmonic tells how far to go up the sin's and cos's
	// we do 1, sin(x), cos(x) ... to sin(n*x) and cos(n*x)
	// where n is the parameter harmonic.
*/
	FunNum = harmonic * 2 + 1;
	if ( FunNum > ciMAXFUNCS )
		FunNum = ciMAXFUNCS;

	MaxY = 0.0;
	MaxX = 0.0;
	MinX = 0.0;
	/*/ load the data */
	for ( i = 0, j = 0; i < ArrayCount; i++ )
	{
		#ifdef PARANOID
		if ( j >= ciMAXPOINTS )
		{
			cerr << "Overflow in Build32" << endl;
			stop_here();
		}
		#endif
		Data[j].X = InArray[i].fX;
		Data[j].Y = InArray[i].fY;
		Data[j].Valid = 1;
		MaxY = ( Data[j].Y > MaxY )? Data[j].Y : MaxY;
		MaxX = ( Data[j].X > MaxX )? Data[j].X : MaxX;
		MinX = ( Data[j].X < MinX )? Data[j].X : MinX;
		j++;
	}
	DataNum = j;


	OffsetY = MaxY/1.9;
	OffsetX = (MaxX + MinX)/2.0;
	if ( OffsetX > cfCNTR_LIM )
	{
#if 0
		if ( verbose )
			cerr << "OffsetX too far " << OffsetX << endl;
#endif
		OffsetX = cfCNTR_LIM;
	}
	else if ( OffsetX < -cfCNTR_LIM )
	{
#if 0
		if ( verbose )
			cerr << "OffsetX too far " << OffsetX << endl;
#endif
		OffsetX = -cfCNTR_LIM;
	}

	polarize();
	/* default retries = 2 so we do the loop 3 times */
	while ( 1 )
	{
		LoadFunctions();
		GramSchmidt();
		solve();

		if ( ++count > retrys )
			break; /* we are done. */
		if ( ! weed(major, count) )
			break; /* we are done. */
	}

	ValidPerCent = ((float)NumEndValidData)/ NumStartValidData;
#if 0
	if ( 0.85 > ValidPerCent )
		cerr << "Less Than 85% " << 100 * ValidPerCent << "%" << endl;
#endif

/*
	// What number to put here?
	// 80% works for all good logs; 70% works with some cat faces and
	// currently passes all known none bad patterns.
	// All known bad ones (1 which we have data for) would be 
	// rejected with 50%
*/
	if ( 0.70 > ValidPerCent )
		return 0; /*/ fail */

	return Fill32(); /* // -1 is ok, 0 on fail */
}

#if 0
#if defined(TESTING)      // produces standalone executable.
#ifndef sparc
#include <alloc.h>        // coreleft()
#include <conio.h>        // getch()
#endif
#include "log_info.h"

POINT_I s32[32];              // defined in scan?.cpp
float gfScaleFactor = 1.0;    // defined in scan?.cpp
LOG_INFO sLog;

int main( int argc, char *argv[] ) {
  POINT_F apsPoint[120];
  POINT_F asGoodCamPtF[120];
  POINT_F sCentroid;
  char zBuf[100];
  int i, j;
  ostrstream buf( zBuf, sizeof zBuf );

  if ( argc < 2 ) {
    cerr << "USAGE: build32 <logs.*>" << endl;
    return 1;
  }

  #if defined(DISPLAY)
  if ( fg_init() == NULL ) {
    cout << "Can't init graphics" << endl;
    return 1;
  }
  fg_set_8x8_font();
  DisplayMode();
  fg_filldisplay(FG_BLUE, FG_MODE_SET, ~0);      // clears screen
  #endif

  InitBuild32();

  ofstream Outfile( "rings.xxx" );
  ofstream DiagData( "diag.dat" );
  //ofstream PointData( "h:\\real.dat" );   // dumps camera points

  for (int iArgCnt = 1; iArgCnt < argc; ++iArgCnt ) {
  #ifdef sparc
    ifstream Infile( argv[iArgCnt] );
  #else
    ifstream Infile( argv[iArgCnt], ios::binary );
  #endif
    if ( !Infile ) {
      #if defined(DISPLAY)
      fg_term();
      #endif
      cerr << "Can't open " << argv[iArgCnt] << endl;
      return 1;
    }

    // loop thru data file building rings.xxx

    while ( Infile ) {
      Infile.seekg( 600, ios::cur );        // skip sRaw data
      Infile.read((char*)&sCentroid, sizeof sCentroid);
	  sCentroid.SwapEnds();
      Infile.read((char*)&apsPoint,  sizeof apsPoint[0] * 120);
	  for ( i = 0; i < 120; i++ )
		apsPoint[i].SwapEnds();
	  #ifdef sparc
      if ( ! Infile )
		break;
	  #endif
      Infile.seekg( sizeof s32, ios::cur );  // skip saved s32 array
      //for ( int iX = 0; iX < 120; ++iX )
      //  PointData << apsPoint[iX] << endl;
      int iGoodPoints = LoadPoints( asGoodCamPtF, apsPoint );
      if ( Build32( asGoodCamPtF, iGoodPoints )) {
		  #ifdef DISPLAY
		  fg_line_t Line;
		  for ( j = 0; j < 32; j += 8 )
			  for ( i = 1; i < 8; i++ )
			  {
				fg_make_line ( Line, xfudge(s32[j + i-1].iX/1000.0),
					yfudge(s32[j + i-1].iY/1000.0),
					xfudge(s32[j + i].iX/1000.0),
					yfudge(s32[j + i].iY/1000.0) );
				fg_drawline( FG_GREEN, FG_MODE_SET, ~0, FG_LINE_SOLID, Line );
			  }
		  #endif
	  #ifdef sparc
			Outfile << NumStartValidData << " " << NumEndValidData
				<< " " << float(NumEndValidData)/NumStartValidData << " ";
			for ( i = 0; i < 5; i++ )
				Outfile << Coeff[i] << " ";
			Outfile << endl;
	  #else
        Outfile << sCentroid << endl;
        for ( int iI = 0; iI < ciPOINTS_PER_RING; ++iI )
          Outfile << s32[iI].iX << " " << s32[iI].iY << endl;
        buf << argv[iArgCnt] << " - good ring" << ends;
	  #endif
      } else
        buf << "Bad ring" << ends;
      center_string(30, buf.str());
      buf.seekp(0);         // reset buf pointer
      buf << coreleft() << ends;
      center_string(50, buf.str());
      buf.seekp(0);         // reset buf pointer
      buf << "Width = " << s32[0].iX - s32[16].iX <<
        " Height = " << s32[8].iY - s32[24].iY << ends;
      center_string(50, buf.str());
      buf.seekp(0);         // reset buf pointer
      delay(500);

      DiagData << s32[0].iX - s32[16].iX << ","
        << s32[8].iY - s32[24].iY << endl;

	#ifdef sparc
	if ( ValidPerCent < 0.85 )
		stop_here();
	#endif
      if (kbhit()) {
        char ch = getch();
        if ( ch == 27 ) {
          ungetch( ch );
          break;
        }
      }
      if ( sLog.iLeadingEdge == 240 )
        break;
    }         // end of while Infile
  }               // end of wildargs loop
  #if defined(DISPLAY)
  fg_term();
  #endif
  return 0;
}
#endif

#endif
