// Copyright 1995 SF Bellenot
// All Rights Reserved
// --yes it is overkill: bellenot@math.fsu.edu
// skew.cc

#include "bench.h"
#include "skew.h"
#include "skewhelp.h"

//#define PARANOID
//int always_number_1;

Skew::Skew ( x32_LogSurf * ls, int bot, int top, Real off )
{
	log = ls;
	bot_row = bot;
	top_row = top;
	offset = off;
}

#define LogX(i,j) log[i][j].x
#define LogZ(i,j) log[i][j].z
#define LogVector(i,j) log[i][j]

int Skew::top_intersection ( Real intersection_x )
{
	int left, right, mid;

	left = MAX_COLS / 2;
	right = 0;

	// zero offset

	if ( ( LogX(top_row,left) > intersection_x )
		|| ( LogX(top_row,right) < intersection_x ) )
		return FALSE;

	for ( mid = (left + right)/2; (left - right) > 1; mid = (left + right)/2 )
	{
		if ( LogX(top_row,mid) <= intersection_x )
			left = mid;
		else
			right = mid;
	}

	if ( LogX(top_row,left) == intersection_x )
	{
		top_position = LogVector(top_row,left);
		top_right = left;
	}
	else if ( LogX(top_row,right) == intersection_x )
	{
		top_position = LogVector(top_row,right);
		top_right = right;
	}
	else
	{
		top_position.x = intersection_x;
		top_position.z = ZPointOnLineX ( intersection_x,
			LogVector(top_row,left),
			LogVector(top_row,right) );
		top_right = right;
	}

	//length = 0.0;
	return TRUE;
}


int Skew::bot_intersection ( Real intersection_x )
{
	int left, right, place, mid;

	left = MAX_COLS / 2;
	place = MAX_COLS;
	right = 0;

	// compute offset
	intersection_x += offset;

	if ( ( LogX(bot_row,left) > intersection_x )
		|| ( LogX(bot_row,right) < intersection_x ) )
		return FALSE;

	for ( mid = (left + place)/2; (place - left) > 1; mid = (left + place)/2 )
	{
		if ( LogX(bot_row,mid) <= intersection_x )
			left = mid;
		else
			place = mid;

	}

	right = (place == MAX_COLS) ? 0 : place;

	if ( LogX(bot_row,left) == intersection_x )
	{
		bot_position = LogVector(bot_row,left);
		bot_left = left;
	}
	else if ( LogX(bot_row,right) == intersection_x )
	{
		bot_position = LogVector(bot_row,right);
		bot_left = right;
	}
	else
	{
		bot_position.x = intersection_x;
		bot_position.z = ZPointOnLineX ( intersection_x,
			LogVector(bot_row,left),
			LogVector(bot_row,right) );
		bot_left = left;
	}

	//length = 0.0;
	return TRUE;
}

int Skew::intersection ( Real intersection_x )
{
	if ( ! this->top_intersection ( intersection_x ) )
		return FALSE;
	else if ( ! this->bot_intersection ( intersection_x ) )
		return FALSE;
	else
	{
		//length = top_position.z - bot_position.z;
		return TRUE;
	}
}

#define BOT_FLAG  0x01
#define TOP_FLAG  0x02
#define BOTH_FLAG  (BOT_FLAG | TOP_FLAG)
#define START 	1
#define FAR		2
#define SHORT	3

int Skew::shift_calc ( Real new_length )
{
	int pt_flag;
	int mv_flag = START;
	int bot_temp;
	Real new_x, saved_x, saved_length, test_length;

	bot_left = 0;
	top_right = 0;

	bot_position = LogVector(bot_row,bot_left);
	top_position = LogVector(top_row,top_right);
	top_position.x += offset;

	while ( top_position.z - bot_position.z < new_length )
	{
		if ( bot_position.x < top_position.x )
		{
			top_right++;
			// bot_left == 0 is like bot_left = 32
			if ( (top_right > MAX_COLS/4) && bot_left
					&& (bot_left < 3 * MAX_COLS/4) )
				return FALSE;
			top_position = LogVector(top_row,top_right);
			top_position.x += offset;
		}
		else
		{
			bot_left = (bot_left == 0 ) ? MAX_COLS - 1 : bot_left - 1;
			if ( (bot_left < 3 * MAX_COLS/4) && (top_right > MAX_COLS/4) )
				return FALSE;
			bot_position = LogVector(bot_row,bot_left);
		}
	}

	if ( bot_left == 0 )
		while ( bot_position.x < top_position.x )
		{
			top_right++;
			top_position = LogVector(top_row,top_right);
			top_position.x += offset;
		}

	if ( top_right == 0 )
		while ( top_position.x < bot_position.x )
		{
			bot_left = (bot_left == 0 ) ? MAX_COLS - 1 : bot_left - 1;
			bot_position = LogVector(bot_row,bot_left);
		}

compute_alt:

	if ( bot_position.x < top_position.x )
	{
		bot_position.x = top_position.x;
		bot_position.z = ZPointOnLineX ( top_position.x,
			LogVector ( bot_row, bot_left ),
			LogVector ( bot_row,
				(bot_left == MAX_COLS - 1)? 0 : (bot_left + 1) ) );
		pt_flag = TOP_FLAG;
	}
	else if ( bot_position.x > top_position.x )
	{
		top_position.x = bot_position.x;
		top_position.z = ZPointOnLineX ( bot_position.x - offset,
			LogVector ( top_row, top_right ),
			LogVector ( top_row, top_right - 1 ) );
		pt_flag = BOT_FLAG;
	}
	else
	{
		pt_flag = BOTH_FLAG;
	}

	test_length = top_position.z - bot_position.z;

	if ( new_length < test_length ) // gone too far
	{
		if ( ( (bot_left == 0) && (pt_flag & BOT_FLAG) )
				|| ( (top_right == 0) && (pt_flag & TOP_FLAG) ) ) 
		{
			if ( top_right )
				top_right--;

			goto exit_fix_offset;
		}
		if ( mv_flag == SHORT )
		{
			goto exit_fix_all;
		}
		else
		{
			saved_x = top_position.x;
			saved_length = test_length;
			mv_flag = FAR;

			bot_temp = (bot_left == MAX_COLS - 1) ? 0: bot_left + 1;

			if ( LogX(bot_row,bot_temp) < LogX(top_row,top_right - 1) )
			{
				bot_left = bot_temp;
				bot_position = LogVector(bot_row,bot_left);
				if ( pt_flag == BOT_FLAG )
				{
					top_position = LogVector(top_row,top_right);
					top_position.x += offset;
				}
			}
			else
			{
				top_right--;
				top_position = LogVector(top_row,top_right);
				top_position.x += offset;
				if ( pt_flag == TOP_FLAG )
				{
					bot_position = LogVector(bot_row,bot_left);
				}
			}
			goto compute_alt;
		}
	}
	else if ( new_length > test_length ) // short
	{
		if ( mv_flag == FAR )
		{
			goto exit_fix_all;
		}
		else
		{
			mv_flag = SHORT;
			saved_length = test_length;
			saved_x = top_position.x;
			if ( pt_flag & BOT_FLAG )
			{
				bot_left = (bot_left) ? bot_left - 1: MAX_COLS - 1;
				if ( (bot_left < 3 * MAX_COLS/4) && (top_right > MAX_COLS/4) )
					return FALSE;
				bot_position = LogVector(bot_row,bot_left);
				top_position = LogVector(top_row,top_right);
				top_position.x += offset;
			}
			else
			{
				top_right++;
				// bot_left == 0 is like bot_left == 32
				if ( (top_right > MAX_COLS/4) && bot_left
						&& (bot_left < 3 * MAX_COLS/4) )
					return FALSE;
				top_position = LogVector(top_row,top_right);
				top_position.x += offset;
				bot_position = LogVector(bot_row,bot_left);
			}
			goto compute_alt;
		}
	}
	else if ( new_length == test_length ) // exact
	{
		goto exit_fix_offset;
	}

exit_fix_all:
	new_x = XGivenTwoLen ( new_length, saved_length,
		test_length, saved_x, top_position.x );

	while ( LogX(bot_row,bot_left) > new_x )
		bot_left = bot_left ? bot_left - 1: MAX_COLS - 1;

	while ( LogX(top_row,top_right) < new_x - offset )
		top_right--;
		
	bot_position.x = new_x;
	bot_position.z = ZPointOnLineX ( new_x,
		LogVector(bot_row,bot_left),
		LogVector(bot_row,(bot_left==MAX_COLS-1)? 0: bot_left + 1) );
	top_position.x = new_x - offset;
	top_position.z = ZPointOnLineX ( top_position.x,
		LogVector(top_row,top_right+1),
		LogVector(top_row,top_right) );
	return TRUE;
	
exit_fix_offset:
	top_position.x -= offset;
	return TRUE;
}

int Skew::limit_shift_calc ( Real new_length, Real Topz, Real Botz )
{
	Real NewHeight, OldHeight;
	int i;

#ifdef PARANOID
	// assumes skew can't overhang both lines
	if ( new_length > Topz - Botz )
		if ( (new_length - Topz + Botz) > 0.001 )
		{
			//printf ( "error in limited shift calc parameters\n" );
		}
#endif

	if ( ! this->shift_calc ( new_length ) )
		return FALSE;
	
	if ( bot_position.z < Botz )
		goto overhangs_bottom;
	else if ( top_position.z > Topz )
		goto overhangs_top;
	else
		return TRUE;

overhangs_bottom:
	OldHeight = LogZ(top_row,top_right) - Botz;
	i = top_right + 1;

	for ( ;; )
	{
		NewHeight = LogZ(top_row,i) - Botz;

		if ( NewHeight < new_length )
		{
			top_right = i;
			OldHeight = NewHeight;
			i++;

			if ( i > MAX_COLS/2 )
				return FALSE;
		}
		else if ( NewHeight == new_length )
		{
			top_position = LogVector(top_row,i);
			bot_position.x = top_position.x + offset;
			bot_position.z = Botz;
			// bot_postion.y is set in shift_calc
			//length = top_position.z - Botz;

			return TRUE;	
		}
		else // it's between top_right and i
		{
			//compute top_position
			// this is easier than the general case
			Real span, alpha, beta;
			Vector L, R;

			span = NewHeight - OldHeight;
			alpha = (new_length - OldHeight )/ span;
			beta = (NewHeight - new_length )/ span;

			L = scalar_mult ( alpha, LogVector(top_row,i) );
			R = scalar_mult ( beta,
				LogVector(top_row,top_right) );
			top_position = v_add ( L, R );

			bot_position.x = top_position.x + offset;
			bot_position.z = Botz;
			// bot_postion.y is set in shift_calc
			//length = top_position.z - Botz;

			return TRUE;	
		}
	}

overhangs_top:
	// must handle the case bot_left == 0
	if ( bot_left == 0 )
		bot_left = MAX_COLS - 1;

	i = bot_left + 1;
	i = (i == MAX_COLS) ? 0 : i;
	OldHeight = Topz - LogZ(bot_row,i);

	for ( ;; )
	{
		NewHeight = Topz - LogZ(bot_row,bot_left);

		if ( NewHeight < new_length )
		{
			i = bot_left;
			bot_left--;
			OldHeight = NewHeight;

			if ( bot_left < 3 * MAX_COLS/4 )
				return FALSE;
		}
		else if ( NewHeight == new_length )
		{
			bot_position = LogVector(bot_row,bot_left);
			top_position.x = bot_position.x - offset;
			top_position.z = Topz;
			// top_postion.y is set in shift_calc
			//length = Topz - bot_position.z;

			return TRUE;	
		}
		else // it's between bot_left and i
		{
			//compute top_position
			// this is easier than the general case
			Real span, alpha, beta;
			Vector L, R;

			span = NewHeight - OldHeight;
			alpha = (new_length - OldHeight )/ span;
			beta = (NewHeight - new_length )/ span;

// reverse of overhangs bottom case above
			R = scalar_mult ( beta, LogVector(bot_row,i) );
			L = scalar_mult ( alpha,
				LogVector(bot_row,bot_left) );
			bot_position = v_add ( L, R );

			top_position.x = bot_position.x - offset;
			top_position.z = Topz;
			// top_postion.y is set in shift_calc
			//length = Topz - bot_position.z;

			return TRUE;	
		}
	}
}

