// Copyright 1995 SF Bellenot
// All Rights Reserved
// --yes it is overkill: bellenot@math.fsu.edu
//
//slice.c++
//
#include "bench.h"
extern int LogMaxI;

#define LogX(i,j) log[i][j].x
#define LogZ(i,j) log[i][j].z
#define LogVector(i,j) log[i][j]

void BananaSlice(x32_LogSurf *log, Real x, Real Top[], Real Bot[], int Status[])
{
	int i, current_j, real_j;
	int top_right, bot_right;
	int otherCol;

/*
 *      For each row i determine top_right and bot_right so that
 *      log[i][top_right + 1].x < x <=  log[i][top_right].x
 *	      and if real_j = MAX_COLS - bot_right
 *      log[i][real_j - 1].x < x <=  log[i][real_j mod MAX_COLS].x
 *
 *      both top_right and bot_right range from -1 to MAX_COLS/2
 *      -1 meaning wide right and MAX_COLS/2 meaning wide left
 *
 *      The top_right uses the top of surface, while bot_right uses
 *      the bottom of the log surface in sort of a backwards column order.
 *
 *      At each stage we determine if our line is wide, above, below or inside
 *      the surface of the log.
 */


	//initial row(0) and column(start at middle)
	// binary search would be faster
	// maybe not, each binary search requires 4 (2^4 = 16) maybe more
	// expected search is half way which is 4 or 5
	// perhaps a toss up

	 i = 0;
	current_j = MAX_COLS / 4;

/*
 *      finite state machine.
 */
start_row:
	if(current_j == -1)
	{
		current_j = 0;
	}
	if(LogX(i, current_j) >= x)
	{
		goto point_higher_equal;
	}
	else
	{
		goto point_lower;
	}

start_bot:
	if(current_j == -1)
	{
		current_j = 0;
	}
	real_j = (current_j == 0) ? 0 : MAX_COLS - current_j;
	if(LogX(i, real_j) >= x)
	{
		goto bot_point_higher_equal;
	}
	else
	{
		goto bot_point_lower;
	}

point_higher_equal://LogX(i, current_j) >= x
	if(current_j == MAX_COLS / 2)
	{
		top_right = current_j;
		Status[i] = WIDE_LEFT;
		goto increment_row;
	}
	else
	{
		current_j++;
		if(LogX(i, current_j) >= x)
		{
			goto point_higher_equal;
		}
		else
		{
			top_right = current_j - 1;
			if(top_right < 0)
			{
				Status[i] = WIDE_RIGHT;
				goto increment_row;
			}
			else
			{
				goto top_found;
			}
		}
	}

bot_point_higher_equal://LogX(i, real_j) >= x
	if(current_j == MAX_COLS / 2)
	{
		bot_right = current_j;
		Status[i] = WIDE_LEFT;
		goto increment_row;
	}
	else
	{
		current_j++;
		real_j = (current_j == 0) ? 0 : MAX_COLS - current_j;
		if(LogX(i, real_j) >= x)
		{
			goto bot_point_higher_equal;
		}
		else
		{
			bot_right = current_j - 1;
			if(bot_right < 0)
			{
				Status[i] = WIDE_RIGHT;
				goto increment_row;
			}
			else
			{
				goto found;
			}
		}
	}

point_lower://LogX(i, current_j) < x
	if(current_j == 0)
	{
		top_right = current_j;
		Status[i] = WIDE_RIGHT;
		goto increment_row;
	}
	else
	{
		current_j--;
		if(LogX(i, current_j) >= x)
		{
			top_right = current_j;
			goto top_found;
		}
		else
		{
			goto point_lower;
		}
	}

bot_point_lower://LogX(i, real_j) < x
	if(current_j == 0)
	{
		bot_right = current_j;
		Status[i] = WIDE_RIGHT;
		goto increment_row;
	}
	else
	{
		current_j--;
		real_j = (current_j == 0) ? 0 : MAX_COLS - current_j;
		if(LogX(i, real_j) >= x)
		{
			bot_right = current_j;
			goto found;
		}
		else
		{
			goto bot_point_lower;
		}
	}

top_found:
// top_right is < MAX_COLS/2
	if ( x == LogX( i, top_right + 1 ) )
		Top[i] = LogZ( i, top_right + 1 );
	else
		Top[i] = ZPointOnLineX ( x, LogVector ( i, top_right ),
			LogVector ( i, top_right + 1 ) );
	//Top[i] = slice_distance(x, i, top_right);
	goto start_bot;

found:
	otherCol = (bot_right)? MAX_COLS - bot_right: 0;
	if ( x == LogX ( i, otherCol ) )
		Bot[i] = LogZ ( i, otherCol );
	else
		Bot[i] = ZPointOnLineX ( x, LogVector ( i, MAX_COLS - bot_right - 1 ),
			LogVector ( i, otherCol ) );

	//real_j = MAX_COLS - bot_right;
	//Bot[i] = slice_distance(x, i, real_j - 1);
	Status[i] = NOT_WIDE;

increment_row:
#ifdef CROSS_COUNT
	if(Status[i] != NOT_WIDE)
	{
		if(i == 0)
		{
		}
		else if(Status[i - 1] == NOT_WIDE)
		{
			qnum_cross++;
		}
	}
#endif

	i++;
	if(i <= LogMaxI)
	{
		goto start_row;
	}
#ifdef CROSS_COUNT
	else
	{
		if(Status[LogMaxI] == NOT_WIDE)
		{
			qnum_cross++;
		}
	}
#endif
}
